<?php
/**
 * Created by PhpStorm.
 * @author mihovil.bubnjar
 * @date 15.08.2016
 * @time 11:09
 */

namespace SicoCreditPlus\Components;


use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\AbstractQuery;
use Shopware\Models\Article\Article;
use Shopware\Models\Article\Detail as ArticleDetail;
use Shopware\Models\Article\Price as ArticlePrice;
use Shopware\Models\Article\Repository as ArticleRepository;
use Shopware\Models\Order\Detail as OrderDetail;
use Shopware\Models\Order\Order;
use Shopware\Models\Order\Status;
use SicoCreditPlus\Lib\Controller\CreditPlusWebshopAPI;
use SicoCreditPlus\Lib\CreditPlusHelper\AbstractShopLogger;
use SicoCreditPlus\Lib\CreditPlusObjects\WebshopContract;
use SicoCreditPlus\Lib\CreditPlusObjects\WebshopCurrency;
use SicoCreditPlus\Lib\CreditPlusObjects\WebshopFinanceArticle;
use SicoCreditPlus\Lib\CreditPlusObjects\WebshopRateTableMonthRow;
use SicoCreditPlus\Lib\CreditPlusObjects\WebshopVoucher;
use SicoCreditPlus\Models\RequestQueue\Repository;
use SicoCreditPlus\Models\RequestQueue\RequestQueue;

class SicoCreditPlusHelper {

	/**
	 * Local static storage for requested Contract Data
	 *
	 * @var WebshopContract[] $_oContractData
	 */
	private static $_oContractData = array();
	/** @var WebshopFinanceArticle $_oFinancingArticleReference */
	private $_oFinancingArticleReference = null;
	/** @var string $_sFinancingMode */
	private $_sFinancingMode = null;
	/** @var CreditPlusWebshopAPI $_oWSApi */
	private static $_oWSApi = null;

	/** @var array $_aCreditCancelStates All Credit States in which a cancel action is possible */
	private $_aCreditCancelStates = array(
		24,
		20,
		21,
		22,
		23,
		25,
	);

	/** @var array $_aCreditCancelStates All Credit States in which a return action is possible */
	private $_aCreditReturnStates = array(
		24,
		99,
		32
	);

	/** @var Status[] $_aPaymentStatesCache All Payment States which have once been requested */
	private static $_aPaymentStatesCache = array();

	/**
	 * @return CreditPlusWebshopAPI
	 */
	public function getWebshopAPI() {
		if ( self::$_oWSApi !== null ) {
			return self::$_oWSApi;
		}
		$oConfig = Shopware()->Config();
		$aSoapParams = array(
			'soap-user' => $oConfig->getByNamespace('SicoCreditPlus', 'sSoapUser'),
			'soap-pass' => $oConfig->getByNamespace('SicoCreditPlus', 'sSoapPass'),
			'soap-type' => $oConfig->getByNamespace('SicoCreditPlus', 'sSoapType'),
			'wsdl' => $oConfig->getByNamespace('SicoCreditPlus', 'sWSDL'),
		);
		$oWSApi = new CreditPlusWebshopAPI($aSoapParams);
		$oWSApi->setDebugState(true, false);
		$oWSApi->setLogger($this->getLogger());

		return self::$_oWSApi = $oWSApi;
	}

	/**
	 * @return AbstractShopLogger|null
	 */
	public function getLogger() {
		$oConfig = Shopware()->Config();
		if ( $oConfig->getByNamespace('SicoCreditPlus', 'bLogActive') ) {
			$oLogger = new SicoCreditPlusLogger();
			if ( class_exists('Shopware\Components\DependencyInjection\Container') ) {
				$oContainer = Shopware()->Container();
				$oLogger->setContainer($oContainer);
			}
			return $oLogger;
		}
		return null;
	}
	/**
	 * Takes general info about the currency available in the shop and returns an object as required for the API
	 *
	 * @return WebshopCurrency
	 */
	public function getWebshopCurrency() {
		$oCurrency = Shopware()->Shop()->getCurrency();
		$aWSCurrency = array(
			'decimals' => 2,
			'decimalSeparator' => ',',
			'thousandSeparator' => '.',
			'side' => (($oCurrency->getSymbolPosition()==0)?'left':'right'),
			'sign' => $oCurrency->getSymbol(),
			'name' => $oCurrency->getName()
		);
		return WebshopCurrency::fromArray($aWSCurrency);
	}

	/**
	 * @param ArticleDetail $oArticleDetails
	 * @param float $dMinRate
	 * @param float $dOrderPrice
	 *
	 * @return WebshopRateTableMonthRow[]
	 */
	public function getFinancingMonths( $oArticleDetails = null, $dMinRate = null, $dOrderPrice = null ) {
		$dOrderPrice = $this->convertPriceFromString2Float($dOrderPrice);
		/** @var WebshopRateTableMonthRow[] $aReturn */
		$aReturn = array();
		if ( $oArticleDetails == null ) { return $aReturn; }
		if ( $this->isPaymentExcludedByMissingApi() ) { return $aReturn; }
		$oConfig = Shopware()->Config();

		$oWSCurrency = $this->getWebshopCurrency();

		if ( $dOrderPrice === null ) {
			/** @var \Shopware\Models\Article\Price[]|ArrayCollection $aPrices */
			$aPrices = $oArticleDetails->getPrices();
			foreach ( $aPrices as $oPrice ) {
				if ( $oPrice->getFrom() <= 1 ) {
					if ( !$oPrice->getTo() || ($oPrice->getTo() == 'beliebig') || ($oPrice->getTo() >= 1) ) {
						$dOrderPrice = $oPrice->getPrice();
						$oCustomerGroup = $oPrice->getCustomerGroup();
						if ( $oCustomerGroup->getTax() ) {
							$dOrderPrice = $dOrderPrice * (1+$oArticleDetails->getArticle()->getTax()->getTax()/100);
						}
					}
				}
			}
		}
		$dMinBasketPrice = floatval($oConfig->getByNamespace('SicoCreditPlus', 'sMinBasketPrice'));
		$dMaxBasketPrice = floatval($oConfig->getByNamespace('SicoCreditPlus', 'sMaxBasketPrice'));

		if ( $dMinRate === null ) {
			$dMinRate = floatval($oConfig->getByNamespace('SicoCreditPlus','sMinRate'));
		}

		// Product or basket is cheaper than min price => no financing
		if ( $dMinBasketPrice > $dOrderPrice ) {
			return $aReturn;
		}
		// Product or basket is more expensive than max price => no financing
		if ( $dMaxBasketPrice < $dOrderPrice ) {
			return $aReturn;
		}


		$oWSApi = $this->getWebshopAPI();
		$aReturn = $this->getProductSpecificFinancingMonths($oArticleDetails, $dMinRate, $dOrderPrice, $aReturn, $oWSApi);
		$aReturn = $this->getUnspecificFinancingMonths($dMinRate, $dOrderPrice, $aReturn, $oWSApi);
		$aReturn = $this->getCalculatedFinancingMonths($aReturn, $dOrderPrice, $oWSCurrency);
		return $aReturn;
	}

	private function convertPriceFromString2Float($sPrice){
		if(!is_string($sPrice)){
			return $sPrice;
		}

		$sPrice = str_replace(',','.',$sPrice);
		$fPrice = (double)$sPrice;
		return $fPrice;
	}


	/**
	 * @param ArticleDetail $oArticleDetails
	 * @param float $dMinRate
	 * @param float $dOrderPrice
	 * @param WebshopRateTableMonthRow[] $aReturn
	 * @param CreditPlusWebshopAPI $oWSApi
	 * @return WebshopRateTableMonthRow[] $aReturn + cheaper or new options
	 */
	private function getProductSpecificFinancingMonths( $oArticleDetails, $dMinRate, $dOrderPrice, $aReturn, $oWSApi) {
		$oDB = Shopware()->Db();
		$oRes = $oDB->executeQuery('
			SELECT oo.id oxid, oo.sico_months sccp_months, 
				oo.sico_interest sccp_interest, oo.sico_prodcode sccp_prodcode,
				oo.sico_active sccp_active, oo.sico_ratefactor sccp_ratefactor,
				pg.sico_product_type_id sccp_producttypeid,
				pg.sico_product_class_id sccp_productclassid,
				pg.sico_name sccp_productgroup_name
			FROM sico_offered_option oo
				INNER JOIN sico_productgroup pg ON oo.sico_prodgroup = pg.id
				INNER JOIN sico_productgroup_s_articles pga ON pg.id = pga.productgroup_id
				INNER JOIN s_articles a ON pga.article_id = a.id
				INNER JOIN s_articles_details ad ON a.id = ad.articleID
			WHERE
				(oo.sico_active = 1) AND
				(pga.article_id = :article_id)
			ORDER BY oo.sico_months ASC, oo.sico_interest DESC
			',
			array('article_id' => $oArticleDetails->getArticle()->getId())
		);
		return $this->getOptionsFromResultSet($oRes, $dMinRate, $dOrderPrice, $aReturn, $oWSApi, true);
	}

	/**
	 * @param float $dMinRate
	 * @param float $dOrderPrice
	 * @param WebshopRateTableMonthRow[] $aReturn
	 * @param CreditPlusWebshopAPI $oWSApi
	 * @return WebshopRateTableMonthRow[] $aReturn + cheaper or new options
	 */
	private function getUnspecificFinancingMonths( $dMinRate, $dOrderPrice, $aReturn, $oWSApi) {
		$oDB = Shopware()->Db();
		$oRes = $oDB->executeQuery('
			SELECT oo.id oxid, oo.sico_months sccp_months, 
				oo.sico_interest sccp_interest, oo.sico_prodcode sccp_prodcode,
				oo.sico_active sccp_active, oo.sico_ratefactor sccp_ratefactor,
				pg.sico_product_type_id sccp_producttypeid,
				pg.sico_product_class_id sccp_productclassid,
				pg.sico_name sccp_productgroup_name
			FROM sico_offered_option oo
				INNER JOIN sico_productgroup pg ON oo.sico_prodgroup = pg.id
				LEFT JOIN sico_productgroup_s_articles pga ON pg.id = pga.productgroup_id
			WHERE
				(oo.sico_active = 1) AND
				(pga.article_id IS NULL)
			ORDER BY oo.sico_months ASC, oo.sico_interest DESC
			'
		);
		return $this->getOptionsFromResultSet($oRes, $dMinRate, $dOrderPrice, $aReturn, $oWSApi);
	}

	/**
	 * @param \Zend_Db_Statement_Pdo $oRes
	 * @param float $dMinRate
	 * @param float $dOrderPrice
	 * @param WebshopRateTableMonthRow[] $aReturn
	 * @param CreditPlusWebshopAPI $oWSApi
	 * @param bool $bIsProductSpecific
	 * @return WebshopRateTableMonthRow[] $aReturn + cheaper or new options
	 */
	private function getOptionsFromResultSet( $oRes, $dMinRate, $dOrderPrice, $aReturn, $oWSApi, $bIsProductSpecific = false ) {
		if ( $oRes && $oRes->rowCount() ) {
			/** @var \stdClass $oOption */
			while ( $oOption = $oRes->fetchObject() ) {
				$aOption = array(
					'months' => intval($oOption->sccp_months),
					'interestRate' => floatval($oOption->sccp_interest),
					'nominalInterestRate' => $oWSApi->getNominalInterestFromEffectiveInterest(floatval($oOption->sccp_interest)),
					'productCode' => $oOption->sccp_prodcode,
					'productTypeID' => $oOption->sccp_producttypeid,
					'productClassID' => $oOption->sccp_productclassid,
					'productGroupName' => $oOption->sccp_productgroup_name,
					'isProductSpecific' => $bIsProductSpecific
				);
				$dRateFactor = floatval($oOption->sccp_ratefactor);
				if ( $dRateFactor === -1.0 ) {
					$dRateFactor = false;
				}
				$aOption['monthlyRate'] = $oWSApi->getMonthRateByPriceMonthsAndInterest($dOrderPrice,$aOption['months'],$aOption['interestRate'],$dRateFactor);
				if ( $aOption['monthlyRate'] < $dMinRate ) {
					// Skip rates, which are too low
					continue;
				}
				if ( !isset($aReturn[$aOption['months']]) ) {
					// If not yet set, just add
					$aReturn[$aOption['months']] = new WebshopRateTableMonthRow($aOption);
				} elseif ( $aReturn[$aOption['months']]->interestRate > $aOption['interestRate'] ) {
					// If general is cheaper, replace
					$aReturn[$aOption['months']] = new WebshopRateTableMonthRow($aOption);
				} else {
					// Skip
				}
			}
		}
		return $aReturn;
	}

	/**
	 * @param WebshopRateTableMonthRow[] $aReturn
	 * @param float $dOrderPrice
	 * @param WebshopCurrency $oCurrency
	 *
	 * @return WebshopRateTableMonthRow[]
	 */
	public function getCalculatedFinancingMonths( $aReturn, $dOrderPrice, $oCurrency ) {
		// Prepare values for output (Currency sign, number format)
		foreach ( $aReturn as &$oWSOption ) {
			$oWSOption->totalAmount = $oWSOption->monthlyRate*$oWSOption->months;
			// If interest is 0.00 or above
			if ( $oWSOption->interestRate >= 0.00 ) {
				// If total price is lower than base price
				// or if price is higher than base price with 0.00 interest,
				// something must be off on the single payments
				// which will be corrected on the final payment anyway
				if ( ($oWSOption->totalAmount < $dOrderPrice) || (($oWSOption->totalAmount > $dOrderPrice) && ($oWSOption->interestRate == 0.00)) ) {
					$oWSOption->totalAmount = $dOrderPrice;
				}
			}
			$oWSOption->interest = $oWSOption->totalAmount-$dOrderPrice;
			$oWSOption->interestRate = number_format($oWSOption->interestRate, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator).'%';
			$oWSOption->nominalInterestRate = number_format($oWSOption->nominalInterestRate, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator).'%';
			if ( $oCurrency->side && ($oCurrency->side == 'left') ) {
				$oWSOption->interest = $oCurrency->sign.' '.number_format($oWSOption->interest, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator);
				$oWSOption->totalAmount = $oCurrency->sign.' '.number_format($oWSOption->totalAmount, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator);
				$oWSOption->monthlyRate = $oCurrency->sign.' '.number_format($oWSOption->monthlyRate, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator);
			} else {
				$oWSOption->interest = number_format($oWSOption->interest, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator).' '.$oCurrency->sign;
				$oWSOption->totalAmount = number_format($oWSOption->totalAmount, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator).' '.$oCurrency->sign;
				$oWSOption->monthlyRate = number_format($oWSOption->monthlyRate, $oCurrency->decimals, $oCurrency->decimalSeparator, $oCurrency->thousandSeparator).' '.$oCurrency->sign;
			}
		}
		// Sorty by months ascending
		ksort($aReturn);
		return $aReturn;
	}

	/**
	 * @param Order $oOrder
	 * @param ArticleDetail[] $aArticles
	 * @param float $dBasketPrice
	 * @return WebshopFinanceArticle
	 */
	public function getFinancingArticleReference($oOrder = null, $aArticles = null, $dBasketPrice = null) {
		if ( !$oOrder && !$aArticles ) {
			return null;
		}
		$oConfig = Shopware()->Config();
		$dMinRate = floatval($oConfig->getByNamespace('SicoCreditPlus', 'sMinRate'));
		$sBasketStrategy = $oConfig->getByNamespace('SicoCreditPlus', 'sBasketFinancingMode');
		if ( ($this->_sFinancingMode === $sBasketStrategy) && ($this->_oFinancingArticleReference !== null) ) {
			return $this->_oFinancingArticleReference;
		}
		$oWSCurrency = $this->getWebshopCurrency();
		$oWSApi = $this->getWebshopAPI();
		$aWebshopArticles = array();

		if ( $aArticles === null ) {
			$aArticles = $this->getArticlesFromOrder($oOrder);
		}
		foreach ( $aArticles as $iKey => $oArticleDetail ) {
			$dPrice = 0.00;
			$dAmount = 1;
			$dOrderPrice = ($dBasketPrice!=null)?$dBasketPrice:0.00;
			if ( $oOrder ) {
				/** @var OrderDetail $oOrderDetail */
				$oOrderDetail = $oOrder->getDetails()->get($iKey);
				$dPrice = $oOrderDetail->getPrice();
				$dAmount = $oOrderDetail->getQuantity();
				$dOrderPrice = $oOrder->getInvoiceAmount();
			} else {
				/** @var \Shopware\Models\Article\Price[] $aPrices */
				$aPrices = $oArticleDetail->getPrices();
				foreach ( $aPrices as $oPrice ) {
					if ( $oPrice->getFrom() <= $dAmount ) {
						if ( !$oPrice->getTo() || ($oPrice->getTo() === 'beliebig') || ($oPrice->getTo() > $dAmount) ) {
							$dPrice = $oPrice->getPrice();
							$dOrderPrice = $dPrice;
						}
					}
				}
			}
			$aWebshopArticle = array(
				'id' => $iKey,
				'amount' => $dAmount,
				'unitprice' => $dPrice,
				'isProductSpecific' => false
			);
			// Skip free samples (e.g. part list items after exploding)
			if ( $dPrice < 0.01 ) {
				continue;
			}
			$aMonths = $this->getFinancingMonths($oArticleDetail, $dMinRate, $dOrderPrice);
			if ( !$aMonths ) {
				// One of the articles has no financing options, so this will not be offered for the order
				// Return null and handle it in the calling function
				return null;
			}
			foreach ( $aMonths as $oMonthRow ) {
				$aWebshopArticle['isProductSpecific'] = $aWebshopArticle['isProductSpecific'] || $oMonthRow->isProductSpecific;
				$dInterestRate = $oWSApi->retrieveFloatFromFormattedInterestRate($oMonthRow->interestRate, $oWSCurrency);
				if ( !isset($aWebshopArticle['cheapestInterestRate']) || ($dInterestRate < $aWebshopArticle['cheapestInterestRate']) ) {
					$aWebshopArticle['cheapestInterestRate'] = $dInterestRate;
					$aWebshopArticle['productTypeID'] = $oMonthRow->productTypeID;
					$aWebshopArticle['productClassID'] = $oMonthRow->productClassID;
				}
				if ( !isset($aWebshopArticle['mostExpensiveInterestRate']) || ($dInterestRate > $aWebshopArticle['mostExpensiveInterestRate']) ) {
					$aWebshopArticle['mostExpensiveInterestRate'] = $dInterestRate;
				}
				if ( $oMonthRow->isProductSpecific && ( $aWebshopArticle['cheapestInterestRate'] == $dInterestRate ) ) {
					$aWebshopArticle['productTypeID'] = $oMonthRow->productTypeID;
					$aWebshopArticle['productClassID'] = $oMonthRow->productClassID;
				}
				$aWebshopArticle['aMonthRows'][(int)$oMonthRow->months] = $dInterestRate;
			}
			$oWebshopArticle = new WebshopFinanceArticle($aWebshopArticle);
			$aWebshopArticles[$iKey] = $oWebshopArticle;
		}
		$oWebshopArticle = $oWSApi->getFinancingArticleReference($aWebshopArticles, $sBasketStrategy);
		$this->_oFinancingArticleReference = $oWebshopArticle;
		return $this->_oFinancingArticleReference;
	}


	/**
	 * @param Order $oOrder
	 *
	 * @return ArticleDetail[]
	 */
	private function getArticlesFromOrder( $oOrder ) {
		/** @var ArticleRepository $oArticleRepository */
		$oArticleRepository = Shopware()->Models()->getRepository('\Shopware\Models\Article\Article');
		$aArticles = array();
		/** @var OrderDetail[] $aOrderDetails */
		$aOrderDetails = $oOrder->getDetails();
		foreach ( $aOrderDetails as $iKey => $oOrderDetail ) {
			// Skip basket contents which are not normal products
			if ( $oOrderDetail->getMode() != 0 ) {
				continue;
			}
			/** @var \Shopware\Models\Article\Article $oArticle */
			$oArticle = $oArticleRepository->find($oOrderDetail->getArticleId());
			/** @var \Shopware\Models\Article\Detail[] $aArticleDetails */
			$aArticleDetails = $oArticle->getDetails();
			foreach ( $aArticleDetails as $oArticleDetail ) {
				if ($oArticleDetail->getNumber() == $oOrderDetail->getArticleNumber() ) {
					$aArticles[$iKey] = $oArticleDetail;
				}
			}
		}
		return $aArticles;
	}

	/**
	 * @param WebshopRateTableMonthRow[] $aMonths
	 * @return float
	 */
	public function getCheapestInterestRate( $aMonths ) {
		$dCheapestInterestRate = null;
		foreach ( $aMonths as $oMonth ) {
			$dEffInterest = floatval(str_replace(array(',','%'),array('.',''),$oMonth->interestRate));
			if ( ( $dCheapestInterestRate === null ) || ($dEffInterest < $dCheapestInterestRate) ) {
				$dCheapestInterestRate = $dEffInterest;
			}
		}
		return $dCheapestInterestRate;
	}

	/**
	 * @param WebshopRateTableMonthRow[] $aMonths
	 * @return WebshopRateTableMonthRow
	 */
	public function getCheapestRate( $aMonths ) {
		/** @var WebshopRateTableMonthRow $aCheapestRate */
		$aCheapestRate = null;
		foreach ( $aMonths as $oMonth ) {
			$dRate = $this->getNumberFromFromattedValue($oMonth->monthlyRate);
			if ( ( $aCheapestRate === null ) || ($dRate < $this->getNumberFromFromattedValue($aCheapestRate->monthlyRate)) ) {
				$aCheapestRate = $oMonth;
			}
		}
		return $aCheapestRate;
	}

	/**
	 * Filter out non numeric values like " " or €
	 * Converting € 40.412,41 to floatval(40412.41)
	 * @param string $sRate
	 * @return float
	 */
	public function getNumberFromFromattedValue( $sRate ) {
		$sMonthRate = '';
		// Filter out non numeric values like " " or €
		// € 40.412,41 is converted to 40.412,41
		for ( $iPos = 0 ; $iPos < strlen($sRate) ; $iPos++ ) {
			$sSubString = substr($sRate,$iPos,1);
			if ( preg_match('/[0-9,.]/',$sSubString) ) {
				$sMonthRate .= $sSubString;
			}
		}
		$aRate = explode(',', $sMonthRate);
		// 40.412,41 is converted to 40412.41
		if ( (count($aRate) == 2) && (strlen($aRate[count($aRate)]) <= 2) ) {
			$aBigNumber = explode('.',$aRate[0]);
			$sMonthRate = implode('',$aBigNumber).'.'.$aRate[1];
			// 40,412.41 is converted to 40412.41
		} else {
			$aRate = explode('.', $sRate);
			$sMonthRate = implode('',explode(',',$aRate[0])).'.'.$aRate[1];
		}
		// And now we have a number...
		$dRate = floatval($sMonthRate);
		return $dRate;
	}

	/**
	 * Requests payment data from CreditPlus and returns $oOrder with dynamically declared cpContractData
	 *
	 * @param Order $oOrder Order object
	 * @return WebshopContract Data saved at CreditPlus
	 */
	public function getContractData( &$oOrder ) {
		$sDealerOrderNumber = $oOrder->getTransactionId();
		$sOrderNumber = $oOrder->getNumber();
		if ( self::$_oContractData[$sDealerOrderNumber] !== null ) {
			return self::$_oContractData[$sDealerOrderNumber];
		}
		if ( self::$_oContractData[$sOrderNumber] !== null ) {
			return self::$_oContractData[$sOrderNumber];
		}
		if ( ($oOrder->getPayment()->getName() !== 'sccp_financing') || ($sDealerOrderNumber == '') ) {
			return self::$_oContractData[$sDealerOrderNumber] = new WebshopContract();
		}
		// Take the multi getter and use only the expected result
		$aWebshopContracts = $this->getContractDataMultiple(array($sDealerOrderNumber, $sOrderNumber));

		return isset($aWebshopContracts[$sDealerOrderNumber])?$aWebshopContracts[$sDealerOrderNumber]:$aWebshopContracts[$sOrderNumber];
	}

	/**
	 * Requests multiple WebshopContracts by sending the transaction IDs through the API
	 * Utilizes local caching per item if same item is requested more than once.
	 *
	 * @param string[] $aTransactionIDs
	 * @return WebshopContract[] Array of WebshopContracts with TransactionID as its index
	 */
	public function getContractDataMultiple( $aTransactionIDs ) {
		$aReturn = array();
		if ( !$aTransactionIDs ) {
			return array();
		}
		$aKeys = array();
		// Don't refetch already fetched data
		// Empty TransactionID == empty WebshopContract
		foreach ( $aTransactionIDs as $iKey => $sTransactionID ) {
			if ( $sTransactionID == '' ) {
				self::$_oContractData[$sTransactionID] = new WebshopContract();
				$aReturn[$sTransactionID] = self::$_oContractData[$sTransactionID];
				unset($aTransactionIDs[$iKey]);
			} elseif ( self::$_oContractData[$sTransactionID] ) {
				$aReturn[$sTransactionID] = self::$_oContractData[$sTransactionID];
				unset($aTransactionIDs[$iKey]);
			} else {
				$aKeys[$sTransactionID] = $iKey;
			}
		}
		// If requests needs to be made, do it
		if ( $aTransactionIDs ) {
			$oConfig = Shopware()->Config();
			$oWSApi = $this->getWebshopAPI();
			$oCreditOfferResponse = null;
			if ( $oWSApi->isApiCallable() ) {
				$oCreditOfferResponse = $oWSApi->getContractsCPWebshop( array('dealerOrderNumber' => $aTransactionIDs, 'dealerNumber' => $oConfig->getByNamespace('SicoCreditPlus', 'sCPDealer') ) );
			}
			if ( !is_object($oCreditOfferResponse) ) {
				$oCreditOfferResponse = new \stdClass();
				$oCreditOfferResponse->return = array( 0 => false );
			}

			if ( !is_array($oCreditOfferResponse->return) ) {
				$oCreditOfferResponse->return = array($oCreditOfferResponse->return);
			}
			// Cache contract data by TransactionID
			foreach ( $oCreditOfferResponse->return as $oCreditOfferResponseSingle ) {
				$sTransactionID = $oCreditOfferResponseSingle->dealerOrderNumber.'';
				self::$_oContractData[$sTransactionID] = WebshopContract::fromObject($oCreditOfferResponseSingle);
				// This is only provided in state 24
				if ( (int)self::$_oContractData[$sTransactionID]->state === 24 ) {
					$oCompleteUrl = $oWSApi->completeOrderUrlCPWebshop($sTransactionID);
					if ( is_object($oCompleteUrl) && $oCompleteUrl && $oCompleteUrl->completeOrderUrl ) {
						self::$_oContractData[$sTransactionID]->completeOrderUrl = $oCompleteUrl->completeOrderUrl.'';
					}
				}
				$aReturn[$sTransactionID] = self::$_oContractData[$sTransactionID];
				unset($aTransactionIDs[$aKeys[$sTransactionID]]);
			}
		}
		// Return WebshopContract[] (using TransactionID as index)
		return $aReturn;
	}

	/**
	 * @param WebshopContract $oContractData
	 * @return array The WebshopContract as array
	 */
	public function convertContractDataToArray( $oContractData ) {
		$aReturn = array();
		$aReturn['addressChanged'] = $oContractData->addressChanged;
		$aReturn['areaCode'] = $oContractData->areaCode;
		$aReturn['cancelationDate'] = $oContractData->cancelationDate;
		$aReturn['cancelationFrom'] = $oContractData->cancelationFrom;
		$aReturn['conditionName'] = $oContractData->conditionName;
		$aReturn['cpReferenceNumber'] = $oContractData->cpReferenceNumber;
		$aReturn['creditDuration'] = $oContractData->creditDuration;
		$aReturn['dealerNumber'] = $oContractData->dealerNumber;
		$aReturn['dealerOrderNumber'] = $oContractData->dealerOrderNumber;
		$aReturn['decisionDate'] = $oContractData->decisionDate;
		$aReturn['deliveryDone'] = $oContractData->deliveryDone;
		$aReturn['finallyApproved'] = $oContractData->finallyApproved;
		$aReturn['firstDecisionDate'] = $oContractData->firstDecisionDate;
		$aReturn['forename'] = $oContractData->forename;
		$aReturn['id'] = $oContractData->id;
		$aReturn['informationRequest'] = $oContractData->informationRequest;
		$aReturn['krebesReference'] = $oContractData->krebesReference;
		$aReturn['loginWebshop'] = $oContractData->loginWebshop;
		$aReturn['name'] = $oContractData->name;
		$aReturn['orderDate'] = $oContractData->orderDate;
		$aReturn['price'] = $oContractData->price;
		$aReturn['state'] = $oContractData->state;
		$aReturn['stateText'] = $oContractData->stateText;
		$aReturn['street'] = $oContractData->street;
		$aReturn['town'] = $oContractData->town;
		$aReturn['voucher'] = array();
		if ( $oContractData->voucher ) {
			foreach ( $oContractData->voucher as $oVoucher ) {
				$aVoucher = array(
					'dealerOrderNumber' => $oVoucher->dealerOrderNumber,
					'agentName' => $oVoucher->agentName,
					'contractId' => $oVoucher->contractId,
					'date' => $oVoucher->date,
					'dealerNumber' => $oVoucher->dealerNumber,
					'done' => $oVoucher->done,
					'id' => $oVoucher->id,
					'number' => $oVoucher->number,
					'value' => $oVoucher->value
				);
				$aReturn['voucher'][] = $aVoucher;
			}
		}
		return $aReturn;
	}

	/**
	 * @param array $aContractData Contract Data provided by convertContractDataToArray function
	 * @see SicoCreditPlusHelper::convertContractDataToArray
	 * @return array Modified contract data
	 */
	public function convertContractDataArrayForDisplay( array $aContractData ) {

		$iStateNum = intval($aContractData['state']);
		$bDeliveryDone = $aContractData['deliveryDone'];
		$bFinallyApproved = $aContractData['finallyApproved'];

		$sHumanReadableState = 'Undefiniert';
		switch ($iStateNum) {
			case 20:
				$sHumanReadableState = 'Offen';
				break;
			case 24:
				$sHumanReadableState = 'Vorläufig Genehmigt';
				if ( $bFinallyApproved ) {
					$sHumanReadableState = 'Lieferfreigabe';
				}
				break;
			case 25:
				$sHumanReadableState = 'Warten auf Unterlagen';
				break;
			case 32:
				$sHumanReadableState = 'Auszahlung';
				break;
			case 92:
				$sHumanReadableState = 'Weich abgelehnt';
				break;
			case 93:
				$sHumanReadableState = 'Abgelehnt';
				break;
			case 95:
				$sHumanReadableState = 'Storniert';
				break;
			case 99:
				$sHumanReadableState = 'Ausgezahlt';
				break;
		}
		if ( $bDeliveryDone ) {
			$sHumanReadableState .= ' und ausgeliefert';
		}
		$aContractData['stateText'] = $sHumanReadableState;

		return $aContractData;
	}

	/**
	 * @param Order $oOrder The order being cancelled
	 * @param string $sUserName The User cancelling the order (E-Mail Address)
	 * @return array Array with errors or null if none
	 */
	public function cancelOrder( $oOrder, $sUserName = 'User requested' ) {
		// Not my payment? Not my problem! :)
		if ( substr($oOrder->getPayment()->getName(),0,5) !== 'sccp_' ) { return null; }
		$aError = array();
		$oLang = Shopware()->Snippets()->getNamespace('frontend/plugins/payment/creditplus');
		// Only orders, which are not yet cancelled are eligible for API Cancel.
		// So we are going through the uncancelled items and refund them as one call.
		$iOrderStateCancelled = 4;
		if ( defined('Shopware\Models\Order\Status::ORDER_STATE_CANCELLED_REJECTED') ) {
			$iOrderStateCancelled = Status::ORDER_STATE_CANCELLED_REJECTED;
		}
		if ( $oOrder->getOrderStatus()->getId() !== $iOrderStateCancelled ) {
			$oConfig = Shopware()->Config();
			$oWSApi = $this->getWebshopAPI();

			$oContractData = $this->getContractData($oOrder);
			$sDealerOrderNumber = $oContractData->dealerOrderNumber;

			// If delivery is not yet done, and the order is in state "pending", cancel is possible
			if ( !$oContractData->deliveryDone && (in_array($oContractData->state,$this->_aCreditCancelStates)) ) {

				$aCancelData = array(
					'dealerOrderNumber' => $sDealerOrderNumber,
					'cancelationFrom' => $sUserName,
					'dealerNumber' => $oConfig->getByNamespace('SicoCreditPlus', 'sCPDealer')
				);

				$oResponse = $oWSApi->cancelOrderCPWebshop($aCancelData);

				if ( is_object($oResponse) && ( property_exists($oResponse,'confirmation') )  && ( property_exists($oResponse->confirmation,'confirmationItems') )  && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
					$iErrorCode = $oResponse->confirmation->confirmationItems->errorCode;
					$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
					$aError = array(
						'sError' => (500+$iErrorCode),
						'sErrorMessage' => $sErrorMessage
					);
					if ( $iErrorCode == 1 ) {
						// Kein Auftrag gefunden
						$bFound = false;
					} elseif ( $iErrorCode == 2 ) {
						// Pflichtfelder fehlen, kann durch Code heraus nicht passieren
						$bFound = false;
					} elseif ( $iErrorCode == 6 ) {
						// Storno im Status Bezahlt kann nicht stattfinden
						$bFound = false;
					} elseif ( $iErrorCode == 14 ) {
						// Irgendein Pflichtfeld sprengt sein Format/Länge
						// dealerOrderNumber > 40
						// cancelationFrom > 27
						$bFound = false;
					} elseif ( $iErrorCode == 15 ) {
						// Auftragsstatus = Unbekannt
						$bFound = false;
					} elseif ( $iErrorCode == 16 ) {
						// Auftragsstatus = Abgelehnt
						$bFound = false;
					} elseif ( $iErrorCode == 17 ) {
						// Auftragsstatus = Storniert
						$bFound = false;
					} elseif ( $iErrorCode == 18 ) {
						// Auftragsstatus = Wird bezahlt
						$bFound = false;
					} else {
						// Keine Ahnung was passiert ist
						$bFound = false;
					}
					if ( !$bFound ) {
						$aError = array(
							'sError' => (500+$iErrorCode),
							'sErrorMessage' => $sErrorMessage
						);
					}
				}
				// If delivery has happened, then only refunds are available. Try cancel first though, then use refunds.
			} elseif ( $oContractData->deliveryDone && (in_array($oContractData->state,$this->_aCreditReturnStates)) ) {


				$aCancelData = array(
					'dealerOrderNumber' => $sDealerOrderNumber,
					'cancelationFrom' => $sUserName,
					'dealerNumber' => $oConfig->getByNamespace('SicoCreditPlus', 'sCPDealer')
				);

				$oResponse = $oWSApi->cancelOrderCPWebshop($aCancelData);

				if ( is_object($oResponse) && ( property_exists($oResponse,'confirmation') )  && ( property_exists($oResponse->confirmation,'confirmationItems') )  && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
					$iErrorCode = $oResponse->confirmation->confirmationItems->errorCode;
					$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
					$aError = array(
						'sError' => (500+$iErrorCode),
						'sErrorMessage' => $sErrorMessage
					);
					if ( $iErrorCode == 1 ) {
						// Kein Auftrag gefunden
					} elseif ( $iErrorCode == 2 ) {
						// Pflichtfelder fehlen, kann durch Code heraus nicht passieren
					} elseif ( $iErrorCode == 6 ) {
						// Storno im Status Bezahlt kann nicht stattfinden
						$aError = $this->returnOrderAfterError($oWSApi, $oContractData, $sUserName, $sDealerOrderNumber);
					} elseif ( $iErrorCode == 14 ) {
						// Irgendein Pflichtfeld sprengt sein Format/Länge
						// dealerOrderNumber > 40
						// cancelationFrom > 27
					} elseif ( $iErrorCode == 15 ) {
						// Auftragsstatus = Unbekannt
						// Auftragsstatus = in Bearbeitung
						// Hier landet die Bestellung auch, wenn diese endgültig genehmigt ist und bereits versendet wurde.
						$aError = $this->returnOrderAfterError($oWSApi, $oContractData, $sUserName, $sDealerOrderNumber);
					} elseif ( $iErrorCode == 16 ) {
						// Auftragsstatus = Abgelehnt
					} elseif ( $iErrorCode == 17 ) {
						// Auftragsstatus = Storniert
					} elseif ( $iErrorCode == 18 ) {
						// Auftragsstatus = Wird bezahlt
						$aError = $this->returnOrderAfterError($oWSApi, $oContractData, $sUserName, $sDealerOrderNumber);
					} else {
						// Keine Ahnung was passiert ist
					}
				}
			} elseif ( ($oContractData->state === '') || ($oContractData->state === 92) || ($oContractData->state === 93) || ($oContractData->state === 95) ) {
				// If it isn't finished on CreditPlus, it is ok to cancel anyway.
				// If it is forbidden to pay with CreditPlus, cancel is also ok.
				$aError = array(
					'sError' => (401),
					'sErrorMessage' => $oLang->get('ErrorCancelNotAvailable')
				);
			} else {
				// Any other state needs no special treatment
			}
		} else {
			$aError = array(
				'sError' => (400),
				'sErrorMessage' => $oLang->get('ErrorCancelAlreadyCancelled')
			);
		}
		if ( !$aError || ($aError['sError'] == 401) ) {
			// If no error or unimportant error
			$oOrderHandler = Shopware()->Modules()->Order();
			$oOrderHandler->setOrderStatus($oOrder->getId(), $iOrderStateCancelled);
		}
		return $aError;
	}

	/**
	 * @param CreditPlusWebshopAPI $oWSApi The Webshop API Object
	 * @param WebshopContract $oContractData The WebshopContract to be cancelled
	 * @param string $sUserName Backend username (for documentation purposes)
	 * @param string $sDealerOrderNumber DealerOrderNumber (CP2016...)
	 *
	 * @return array array(sError => error_code, sErrorMessage => error_text)
	 */
	protected function returnOrderAfterError($oWSApi, $oContractData, $sUserName, $sDealerOrderNumber) {
		// Collect returnable amount (= price - already returned items)
		$dReturnAmount = floatval($oContractData->price);
		$sDate = date('c');
		foreach ( $oContractData->voucher as $oVoucher ) {
			$dReturnAmount -= floatval($oVoucher->value);
		}
		$oVoucher = new WebshopVoucher($dReturnAmount,'Restbetrag',$sDate,$sUserName,false,$oContractData->id,$sDealerOrderNumber,$oContractData->dealerNumber);
		$oWSApi->addReturnProduct($oVoucher);

		$oResponse = $oWSApi->returnProductCPWebshop();

		// Fehlerhandling
		if ( is_object($oResponse) && ( property_exists($oResponse,'confirmation') ) && ( property_exists($oResponse->confirmation,'confirmationItems') )  && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
			$iErrorCode = $oResponse->confirmation->confirmationItems->errorCode;
			$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
			$aError = array(
				'sError' => (500+$iErrorCode),
				'sErrorMessage' => $sErrorMessage
			);

			if ( $iErrorCode == 1 ) {
				// Kein Auftrag gefunden
			} elseif ( $iErrorCode == 2 ) {
				// Pflichtfelder fehlen, kann durch Code heraus nicht passieren
			} elseif ( $iErrorCode == 7 ) {
				// Retoure im Status "Nicht bezahlt" kann nicht stattfinden
				// Nicht möglich, da dieser Aufruf nur im Status Bezahlt gemacht wird
			} elseif ( $iErrorCode == 12 ) {
				// Summe der Retouren übersteigt Restwert des Kreditvertrags
			} elseif ( $iErrorCode == 14 ) {
				// Irgendein Pflichtfeld sprengt sein Format/Länge
				// dealerOrderNumber > 40
				// cancelationFrom > 27
			} else {
				// Keine Ahnung was passiert ist
			}
		} else {
			$aError = array();
		}
		return $aError;
	}

	/**
	 * @param Order $oOrder The order being modified
	 * @param string $sUserName The User cancelling the order (E-Mail Address)
	 * @return array Array with errors or empty if none
	 */
	public function partiallyReturnOrder( $oOrder, $sUserName = 'User requested' ) {

		$aError = array();
		// If it is not our payment type, avoid our code
		if ( (substr($oOrder->getPayment()->getName(), 0, 5) == 'sccp_') && $oOrder->getNumber() && $oOrder->getTransactionId() ) {
			$sDate = date('c');
			$oContractData = $this->getContractData($oOrder);

			// Price = Contract price - vouchers
			$dInvoiceAmount = $oContractData->price;
			if ( $oContractData->voucher ) {
				foreach ( $oContractData->voucher as $oVoucher ) {
					$dInvoiceAmount -= $oVoucher->value;
				}
			}

			// This is already reduced because of the Hook position (after Backend Action)
			$dNewSum = $oOrder->getInvoiceAmount();

			$dReturnAmount = $dInvoiceAmount-$dNewSum;

			try {
				/** @var \Shopware_Plugins_Backend_Auth_Bootstrap $oAuth */
				$oAuth     = Shopware()->Plugins()->Backend()->Auth();
				$oAuthData = $oAuth->checkAuth();
				if ( $oAuthData && $oAuthData->getIdentity() ) {
					/** @var \StdClass $oUser */
					$oUser = $oAuthData->getIdentity();
					$sUserName = $oUser->username;
				}
			} catch (\Exception $oEx) {
				// At least we tried...
			}

			if ( strlen($sUserName) > 27 ) {
				$sUserName = substr($sUserName,0,27);
			}

			$sDate = date('c');
			$oWSApi = $this->getWebshopAPI();

			$aOrderData = array(
				'dealerOrderNumber' => $oContractData->dealerOrderNumber,
				'changeDate' => $sDate,
				'changedBy' => $sUserName,
				'dealerNumber' => $oContractData->dealerNumber,
				// Storno = Restwert auf neue Summe setzen
				'loanAmount' => $dNewSum,
				'cpReferenceNumber' => $oContractData->cpReferenceNumber
			);

			$oResponse = $oWSApi->changeOrderCPWebshop($aOrderData);
			// Fehlerhandling
			if ( is_object($oResponse) && ( property_exists($oResponse,'confirmation') ) && ( property_exists($oResponse->confirmation,'confirmationItems') )  && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
				$iErrorCode = $oResponse->confirmation->confirmationItems->errorCode;
				$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
				$aError = array(
					'sError' => (500+$iErrorCode),
					'sErrorMessage' => $sErrorMessage
				);
				if ( $iErrorCode == 1 ) {
					// Kein Auftrag gefunden
				} elseif ( $iErrorCode == 2 ) {
					// Pflichtfelder fehlen, kann durch Code heraus nicht passieren
				} elseif ( ($iErrorCode == 6) || ($iErrorCode == 18) || ($iErrorCode == 11)  || ($iErrorCode == 15) ) {
					// changeOrder im Status "Bezahlt", "In Bezahlung", "Ausgeliefert" oder "Lieferfreigabe" kann nicht stattfinden
					$oVoucher = new WebshopVoucher($dReturnAmount,'Warenkorbmodifikation',$sDate,$sUserName,false,$oContractData->id,$oContractData->dealerOrderNumber,$oContractData->dealerNumber);
					$oWSApi->addReturnProduct($oVoucher);

					$oResponse = $oWSApi->returnProductCPWebshop();
					// Fehlerhandling
					if ( is_object($oResponse) && ( property_exists($oResponse,'confirmation') ) && ( property_exists($oResponse->confirmation,'confirmationItems') )  && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
						$iErrorCode = $oResponse->confirmation->confirmationItems->errorCode;
						$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
						$aError = array(
							'sError' => (500+$iErrorCode),
							'sErrorMessage' => $sErrorMessage
						);
					} else {
						$aError = array();
					}
				} elseif ( $iErrorCode == 12 ) {
					// Summe der Retouren übersteigt Restwert des Kreditvertrags
				} elseif ( $iErrorCode == 14 ) {
					// Irgendein Pflichtfeld sprengt sein Format/Länge
					// dealerOrderNumber > 40
					// cancelationFrom > 27
				} else {
					// Keine Ahnung was passiert ist
				}
			} else {
				$aError = array();
			}
		}
		return $aError;
	}

	/**
	 * @param Order $oOrder The order being reduced
	 * @param string[] $aArtNum The Article number of the item being removed
	 * @param int[] $aQuantity The amount of the item being removed
	 * @param float[] $aPrice The price of the item being removed
	 * @return string[] Array on error, null on success
	 */
	public function returnOrderArticles( $oOrder, $aArtNum, $aQuantity, $aPrice = null ) {
		$aError = array();
		// If it is not our payment type, avoid our code
		if ( (substr($oOrder->getPayment()->getName(), 0, 5) == 'sccp_') && $oOrder->getNumber() && $oOrder->getTransactionId() ) {
			$sDate = date('c');
			$oContractData = $this->getContractData($oOrder);

			// Invoice Amount is already reduced before this is called
			$dInvoiceAmount = $oOrder->getInvoiceAmount();
			// Set default values for returns
			$oWSApi = $this->getWebshopAPI();
			$sUserName = 'User Requested';
			try {
				/** @var \Shopware_Plugins_Backend_Auth_Bootstrap $oAuth */
				$oAuth     = Shopware()->Plugins()->Backend()->Auth();
				$oAuthData = $oAuth->checkAuth();
				if ( $oAuthData && $oAuthData->getIdentity() ) {
					/** @var \StdClass $oUser */
					$oUser = $oAuthData->getIdentity();
					$sUserName = $oUser->username;
				}
			} catch (\Exception $oEx) {
				// At least we tried...
			}
			if ( strlen($sUserName) > 27 ) {
				$sUserName = substr($sUserName,0,27);
			}

			// Calculate returned amount
			$dTotalReturnAmount = 0.00;
			foreach ( $aArtNum as $iKey => $sArtNum ) {
				if ( $aPrice[$iKey] === null ) {
					$dReturnAmount = 0.00;
					$oBuilder = Shopware()->Models()->createQueryBuilder();
					$oBuilder->select( array( 'article', 'details' ) )
							 ->from( 'Shopware\Models\Article\Details', 'details' )
							 ->leftJoin( 'details.article', 'article' )
							 ->where( 'details.number = :artnum' )
							 ->setParameter( 'artnum', $sArtNum );
					$oQuery = $oBuilder->getQuery();
					/** @var ArticleDetail $oDetail */
					if ( ! ( $oDetail = $oQuery->getResult( AbstractQuery::HYDRATE_OBJECT ) ) ) {
						return array(
							'sError' => 500,
							'sErrorMessage' => 'Artikel nicht gefunden'
						);
					}
					if ( $oDetail->getNumber() == $sArtNum ) {
						/** @var ArticlePrice[]|ArrayCollection $aPrices */
						$aPrices = $oDetail->getPrices();
						foreach ( $aPrices as $oPrice ) {
							if ( $oPrice->getFrom() <= $aQuantity[$iKey] ) {
								if (
									( $oPrice->getTo() == 'beliebig' )
									|| ( $oPrice->getTo() === null )
									|| ( $oPrice->getTo() >= $aQuantity[$iKey] )
								) {
									$dReturnAmount = $oPrice->getPrice();
								}
							}
						}
					}
				} else {
					$dReturnAmount = $aPrice[$iKey];
				}
				$dReturnAmount = $dReturnAmount * $aQuantity[$iKey];
				$dTotalReturnAmount += $dReturnAmount;
				if ( $dReturnAmount > 0.00 ) {
					// Save vouchers for later (only used if changeOrder fails)
					$oVoucher = new WebshopVoucher($dReturnAmount, $sArtNum, $sDate, $sUserName, false, $oContractData->id, $oContractData->dealerOrderNumber, $oContractData->dealerNumber);
					$oWSApi->addReturnProduct($oVoucher);
				}
			}
			if ( $dTotalReturnAmount == 0.00 ) {
				// No return amount == no need for CreditPlus call
				return array();
			}

			// Zuerst versuchen Bestellung zu ändern
			$aOrderData = array(
				'dealerOrderNumber' => $oContractData->dealerOrderNumber,
				'changeDate' => $sDate,
				'changedBy' => $sUserName,
				'dealerNumber' => $oContractData->dealerNumber,
				// Storno = Restwert auf Gesamtsumme-Preis des stornierten Artikels setzen
				'loanAmount' => $dInvoiceAmount,
				'cpReferenceNumber' => $oContractData->cpReferenceNumber
			);

			$oResponse = $oWSApi->changeOrderCPWebshop($aOrderData);
			// Fehlerhandling
			if ( is_object($oResponse) && ( property_exists($oResponse,'confirmation') ) && ( property_exists($oResponse->confirmation,'confirmationItems') )  && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
				$iErrorCode = $oResponse->confirmation->confirmationItems->errorCode;
				$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
				$aError = array(
					'sError' => (500+$iErrorCode),
					'sErrorMessage' => $sErrorMessage
				);
				if ( $iErrorCode == 1 ) {
					// Kein Auftrag gefunden
				} elseif ( $iErrorCode == 2 ) {
					// Pflichtfelder fehlen, kann durch Code heraus nicht passieren
				} elseif ( ($iErrorCode == 6) || ($iErrorCode == 11)  || ($iErrorCode == 18) || ($iErrorCode == 15) ) {
					// Wenn Status = "Bezahlt", "Ausgeliefert", "In Bezahlung" oder "Lieferfreigabe" - Retoure nutzen
					$oResponse = $oWSApi->returnProductCPWebshop();
					if ( is_object($oResponse) && ( property_exists($oResponse,'confirmation') ) && ( property_exists($oResponse->confirmation,'confirmationItems') )  && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
						$iErrorCode = $oResponse->confirmation->confirmationItems->errorCode;
						$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
						$aError = array(
							'sError' => (500+$iErrorCode),
							'sErrorMessage' => $sErrorMessage
						);
					} else {
						$aError = array();
					}
				} elseif ( $iErrorCode == 12 ) {
					// Summe der Retouren übersteigt Restwert des Kreditvertrags
				} elseif ( $iErrorCode == 14 ) {
					// Irgendein Pflichtfeld sprengt sein Format/Länge
					// dealerOrderNumber > 40
					// cancelationFrom > 27
				}  elseif ( $iErrorCode == 17 ) {
					// Auftrag ist bereits storniert.
				} else {
					// Keine Ahnung was passiert ist
				}
				return $aError;
			}
		}
		return $aError;
	}

	/**
	 * Sends order.
	 * Copied from order_main, extended by commitDeliveryCPWebshop
	 * @param Order $oOrder
	 * @return array
	 * @see Shopware_Controllers_Backend_SicoCreditPlusOrder::sendOrderAction()
	 * @see CreditPlusWebshopAPI::commitDeliveryCPWebshop()
	 */
	public function sendOrder($oOrder)
	{
		$aError = array();
		// Only do this for our payment type
		if ( $oOrder->getPayment()->getName() == 'sccp_financing' ) {
			$oConfig = Shopware()->Config();
			$oWSApi = $this->getWebshopAPI();
			$oContractData = $this->getContractData($oOrder);
			$aData = array(
				'dealerNumber' => $oConfig->getByNamespace('SicoCreditPlus', 'sCPDealer'),
				'dealerOrderNumber' => $oContractData->dealerOrderNumber,
				'invoiceNumber' => $oOrder->getNumber(),
				'invoicePrice' => $oOrder->getInvoiceAmount(),
				'deliveryDate' => date('c')
			);
			$oResponse = $oWSApi->commitDeliveryCPWebshop( $aData );
			if ( is_string($oResponse) ) {
				$aError['sError'] = 500;
				$aError['sErrorMessage'] = $oResponse;
			} else {
				if ( (property_exists($oResponse,'confirmation')) && (property_exists($oResponse->confirmation,'isSucceed')) && ( $oResponse->confirmation->isSucceed ) ) {
					// No error
					// $aError['sError'] = 200;
					// $aError['sErrorMessage'] = 'Status geändert';
				} else {
					if ( ( property_exists($oResponse,'confirmation') ) && ( property_exists($oResponse->confirmation,'confirmationItems') ) && ( property_exists($oResponse->confirmation->confirmationItems,'errorCode') ) ) {
						$iErrorCode = intval($oResponse->confirmation->confirmationItems->errorCode);
						$sErrorMessage = $oResponse->confirmation->confirmationItems->errorMessage;
						$aError = array(
							'sError' => (500+$iErrorCode),
							'sErrorMessage' => $sErrorMessage
						);
						if ( $iErrorCode == 1 ) {
							// Kein Auftrag gefunden
						} elseif ( $iErrorCode == 2 ) {
							// Pflichtfelder fehlen, kann durch Code heraus nicht passieren
						} elseif ( $iErrorCode == 3 ) {
							// Rechnungssumme > Kreditsumme
						} elseif ( $iErrorCode == 4 ) {
							// Lieferdatum in der Zukunft
						} elseif ( $iErrorCode == 5 ) {
							// Vertrag ist nicht genehmigt
						} elseif ( $iErrorCode == 6 ) {
							// Vertrag ist bereits ausbezahlt
						} elseif ( $iErrorCode == 10 ) {
							// Vertrag ist nicht finallyApproved
						} elseif ( $iErrorCode == 11 ) {
							// Vertrag steht bereits auf ausgeliefert
						} elseif ( $iErrorCode == 14 ) {
							// Irgendein Pflichtfeld sprengt sein Format/Länge
							// dealerOrderNumber > 40
							// invoiceNumber > 40
						} elseif ( $iErrorCode == 23 ) {
							// Rechnungssumme < Kreditsumme
						} else {
							// Keine Ahnung was passiert ist
						}
					} else {
						$aError['sError'] = 4.99;
						$aError['sErrorMessage'] = 'An Error has been Xzibited (An Error in an Error has occured)';
					}
				}
			}

			if ( !$aError ) {
				$oOrderHandler = Shopware()->Modules()->Order();
				$iOrderState = 7;
				if ( defined('Shopware\Models\Order\Status::ORDER_STATE_COMPLETELY_DELIVERED') ) {
					$iOrderState = Status::ORDER_STATE_COMPLETELY_DELIVERED;
				}
				$oOrderHandler->setOrderStatus($oOrder->getId(), $iOrderState);
				/** @var \Shopware\Models\Order\Repository $oOrderRep */
				$oOrderRep = Shopware()->Models()->getRepository('Shopware\Models\Order\Order');
				if ( method_exists($oOrderRep, 'getBackendOrdersQuery') ) {
					$oQuery = $oOrderRep->getBackendOrdersQuery(array(array('property' => 'orders.id', 'value' => $oOrder->getId())));
				} else {
					$oQuery = $oOrderRep->getOrdersQuery(array(array('property' => 'orders.id', 'value' => $oOrder->getId())));
				}
				$oQuery->setHydrationMode(AbstractQuery::HYDRATE_ARRAY);
				$oPaginator = Shopware()->Models()->createPaginator($oQuery);
				//returns the customer data
				$aOrders = $oPaginator->getIterator()->getArrayCopy();
				$oEventArgs = new \Enlight_Controller_EventArgs(array(
					'aOrder'  => (isset($aOrders[0])?$aOrders[0]:array()),
					'oOrder'  => $oOrder,
				));
				try {
					Shopware()->Events()->notify(
						'SicoCreditPlus_Financing_Status_Changed',
						$oEventArgs
					);
				} catch ( \Enlight_Event_Exception $e ) {
					// Doesn't happen
				}
			}
		}
		return $aError;
	}


	/**
	 * @return bool true if Payment is excluded (either by Session or Basket)
	 */
	public function isPaymentExcluded() {
		$bSccpExcluded = $this->isPaymentExcludedByMissingApi();

		$bSccpExcluded = $bSccpExcluded || $this->isPaymentExcludedBySession();

		$bSccpExcluded = $bSccpExcluded || $this->isPaymentExcludedByBasket();

		$bSccpExcluded = $bSccpExcluded || $this->isPaymentExcludedByArticlesInBasket();

		return $bSccpExcluded;
	}

	/**
	 * If the API is not callable, the payment type should not be offered
	 * @return bool
	 */
	public function isPaymentExcludedByMissingApi() {
		$oWSAPI = $this->getWebshopAPI();
		if ( !$oWSAPI->isApiCallable() ) {
			return true;
		}
		return false;
	}
	/**
	 * Check Session variable for "excluded"
	 * @return bool
	 */
	public function isPaymentExcludedBySession() {
		return Shopware()->Session()->get('bSicoFinancingExcluded', false);
	}

	/**
	 * Check Basket price for "excluded"
	 * @return bool
	 */
	public function isPaymentExcludedByBasket() {
		// Exclude by price
		$oBasketHandler = Shopware()->Modules()->Basket();
		$oConfig = Shopware()->Config();
		$dMaxBasketPrice = floatval($oConfig->getByNamespace('SicoCreditPlus', 'sMaxBasketPrice'));
		$dMinBasketPrice = floatval($oConfig->getByNamespace('SicoCreditPlus', 'sMinBasketPrice'));
		$bRet = true;
		foreach($oBasketHandler->sGetAmount() as $dBasketPrice){
			$bRet = $bRet && (($dBasketPrice < $dMinBasketPrice) || ($dBasketPrice > $dMaxBasketPrice));
		}

		return $bRet;
	}

	/**
	 * Checks the current basket articles for availability of sccp_financing
	 *
	 * @return bool
	 * @throws \Doctrine\ORM\ORMException
	 * @throws \Doctrine\ORM\OptimisticLockException
	 * @throws \Doctrine\ORM\TransactionRequiredException
	 */
	public function isPaymentExcludedByArticlesInBasket() {
		$oBasketHandler = Shopware()->Modules()->Basket();
		if ( method_exists($oBasketHandler, 'sGetBasketData') ) {
			$aBasketData = $oBasketHandler->sGetBasketData();
		} else {
			// Shopware is older than 5.0.0, do the not so speedy variant
			$aBasketData = $oBasketHandler->sGetBasket();
		}
		// Only available if basket is set
		if ( $aBasketData && $aBasketData['content'] ) {
			$oModelManager = Shopware()->Models();
			$dAmount = $aBasketData['Amount'];
			$dAmount = $this->convertPriceFromString2Float($dAmount);
			foreach ( $aBasketData['content'] as $iKey => $aBasketArticle ) {
				// Skip basket contents which are not normal products
				if ( $aBasketArticle['modus'] != 0 ) {
					continue;
				}
				// Skip free samples (e.g. part list items after exploding)
				if ( $aBasketArticle['priceNumeric'] < 0.01 ) {
					continue;
				}
				// pre 5.0.0 format to 5.0.0 format
				if ( !$aBasketArticle['articleDetailId'] && $aBasketArticle['additional_details'] && $aBasketArticle['additional_details']['articleDetailsID'] ) {
					$aBasketArticle['articleDetailId'] = $aBasketArticle['additional_details']['articleDetailsID'];
				}
				// If only articleDetailId is not given, use base Articles Main Detail
				if ( !$aBasketArticle['articleDetailId'] && $aBasketArticle['articleID'] ) {
					/** @var Article $oArticle */
					$oArticle = $oModelManager->find('\Shopware\Models\Article\Article', $aBasketArticle['articleID']);
					$aBasketArticle['articleDetailId'] = $oArticle->getMainDetail()->getId();
				}
				// Product can not be identified, assume blocked for basket
				if ( !$aBasketArticle['articleDetailId'] ) {
					return true;
				}
				/** @var ArticleDetail $oArticleDetail */
				$oArticleDetail = $oModelManager->find('\Shopware\Models\Article\Detail', $aBasketArticle['articleDetailId']);
				$aMonths = $this->getFinancingMonths($oArticleDetail, null, $dAmount);
				// No months == product excluded
				if ( !$aMonths ) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Returns all CreditPlus payment states
	 *
	 * @return array CreditPlus payment states as array
	 */
	public function getCreditPlusPaymentStates() {
		$aStates = array(
			'creditplus_referred' => array(
				// Krebes 20
				'name' => 'creditplus_referred',
				'description' => 'Antrag ausgefüllt',
				'group' => 'payment',
				'position' => 61,
				'mail' => 0
			),
			'creditplus_accepted' => array(
				// Krebes 24
				'name' => 'creditplus_accepted',
				'description' => 'Antrag angenommen',
				'group' => 'payment',
				'position' => 62,
				'mail' => 0
			),
			'creditplus_declined_soft' => array(
				// Krebes 92
				'name' => 'creditplus_declined_soft',
				'description' => 'Weich abgelehnt',
				'group' => 'payment',
				'position' => 71,
				'mail' => 0
			),
			'creditplus_declined_hard' => array(
				// Krebes 93
				'name' => 'creditplus_declined_hard',
				'description' => 'Abgelehnt',
				'group' => 'payment',
				'position' => 72,
				'mail' => 0
			),
			'creditplus_approved' => array(
				// Krebes 24, Approved
				'name' => 'creditplus_approved',
				'description' => 'Lieferfreigabe',
				'group' => 'payment',
				'position' => 73,
				'mail' => 0
			),
			'creditplus_approved_and_dispatched' => array(
				// Krebes 24, Approved, Delivered
				'name' => 'creditplus_approved_and_dispatched',
				'description' => 'Ausgeliefert',
				'group' => 'payment',
				'position' => 74,
				'mail' => 0
			),
			'creditplus_processing_payment' => array(
				// Krebes 32
				'name' => 'creditplus_processing_payment',
				'description' => 'Auszahlung',
				'group' => 'payment',
				'position' => 75,
				'mail' => 0
			),
			'creditplus_paid' => array(
				// Krebes 99
				'name' => 'creditplus_paid',
				'description' => 'Ausgezahlt',
				'group' => 'payment',
				'position' => 76,
				'mail' => 0
			),
			'creditplus_cancelled' => array(
				// Krebes 95
				'name' => 'creditplus_cancelled',
				'description' => 'Storniert',
				'group' => 'payment',
				'position' => 77,
				'mail' => 0
			),
			'creditplus_error' => array(
				// Krebes 94
				'name' => 'creditplus_error',
				'description' => 'Fehler',
				'group' => 'payment',
				'position' => 78,
				'mail' => 0
			),
			'creditplus_docs_received' => array(
				// Krebes 21, No information request
				'name' => 'creditplus_docs_received',
				'description' => 'Warten auf Unterlagen',
				'group' => 'payment',
				'position' => 79,
				'mail' => 0
			),
		);

		return $aStates;
	}

	/**
	 *
	 * @return string[] Array of oldstate => newstate assignments
	 */
	public function getCreditPlusPaymentStateReplacements() {
		$aReplacements = array();
		// These remain in the current form. If the target gets remapped, just add a line after these.
		$aReplacements['creditplus_docs_received_21_no_information'] = 'creditplus_docs_received';
		$aReplacements['creditplus_docs_received_22_no_information'] = 'creditplus_docs_received';
		$aReplacements['creditplus_docs_received_23_no_information'] = 'creditplus_docs_received';
		$aReplacements['creditplus_docs_received_25_no_information'] = 'creditplus_docs_received';
		$aReplacements['creditplus_docs_received_21_information'] = 'creditplus_docs_received';
		$aReplacements['creditplus_docs_received_22_information'] = 'creditplus_docs_received';
		$aReplacements['creditplus_docs_received_23_information'] = 'creditplus_docs_received';
		$aReplacements['creditplus_docs_received_25_information'] = 'creditplus_docs_received';

		// Clean up replacements to newest id, if one was replaced, rerun replacing for the entire array.
		$bReplaced = true;
		while ( $bReplaced ) {
			$bReplaced = false;
			foreach ( $aReplacements as $sKey => $sValue ) {
				if ( isset($aReplacements[$sValue]) ) {
					$aReplacements[$sKey] = $aReplacements[$sValue];
					$bReplaced = true;
				}
			}
		}

		return $aReplacements;
	}
	/**
	 * @param int $iStateNumber State in Krebes
	 * @param bool $bFinallyApproved Whether this request has been approved by CreditPlus
	 * @param bool $bDeliveryDone Whether the shop has already sent out the goods
	 * @param bool $bRequestInformation Whether or not additional information is requested
	 *
	 * @return string Name of the state for the given parameters
	 */
	public function getCreditPlusPaymentState( $iStateNumber, $bFinallyApproved = false, $bDeliveryDone = false, $bRequestInformation = false ) {
		$sStatus = 'review_necessary';
		if ( $iStateNumber == 20 ) {
			$sStatus = 'creditplus_referred';
		} elseif ( $iStateNumber == 24 ) {
			$sStatus = 'creditplus_accepted';
			if ( $bFinallyApproved == false ) {
				// Already set above
			} elseif ( ($bFinallyApproved == true) && ($bDeliveryDone == true) ) {
				$sStatus = 'creditplus_approved_and_dispatched';
			} elseif ( $bFinallyApproved == true ) {
				$sStatus = 'creditplus_approved';
			}
		} elseif ( $iStateNumber == 21 ) {
			$sStatus = 'creditplus_docs_received';
		} elseif ( $iStateNumber == 22 ) {
			$sStatus = 'creditplus_docs_received';
		} elseif ( $iStateNumber == 23 ) {
			$sStatus = 'creditplus_docs_received';
		} elseif ( $iStateNumber == 25 ) {
			$sStatus = 'creditplus_docs_received';
		} elseif ( $iStateNumber == 32 ) {
			$sStatus = 'creditplus_processing_payment';
		} elseif ( $iStateNumber == 92 ) {
			$sStatus = 'creditplus_declined_soft';
		} elseif ( $iStateNumber == 93 ) {
			$sStatus = 'creditplus_declined_hard';
		} elseif ( $iStateNumber == 99 ) {
			$sStatus = 'creditplus_paid';
		} elseif ( $iStateNumber == 95 ) {
			$sStatus = 'creditplus_cancelled';
		} elseif ( $iStateNumber == 94 ) {
			$sStatus = 'creditplus_error';
		}
		return $sStatus;
	}

	/**
	 * @param string $sPaymentSate
	 *
	 * @return Status
	 */
	public function getPaymentStateByName( $sPaymentSate ) {
		if ( !isset(self::$_aPaymentStatesCache[$sPaymentSate]) ) {
			$oStateRepository = Shopware()->Models()->getRepository('Shopware\Models\Order\Status');
			$aSearchBy = array('name' => $sPaymentSate);
			// 4.3.x code, weil "name" noch nicht im Model existiert
			if ( !property_exists('Shopware\Models\Order\Status', 'name') ) {
				$oSelect = Shopware()->Db()->query("SELECT id FROM s_core_states WHERE name = :name", $aSearchBy);
				$iID = $oSelect->fetchColumn(0);
				$aSearchBy = array('id' => $iID);
			}

			$oPaymentState = $oStateRepository->findOneBy($aSearchBy);
			if ( $oPaymentState ) {
				self::$_aPaymentStatesCache[$sPaymentSate] = $oPaymentState;
			} else {
				self::$_aPaymentStatesCache[$sPaymentSate] = null;
			}
		}
		return self::$_aPaymentStatesCache[$sPaymentSate];
	}


	/**
	 * Triggers Change Mail like in Bootstrap
	 * The code here is taken mostly from Shopware_Controllers_Backend_Order
	 *
	 * @see Shopware_Plugins_Frontend_SicoCreditPlus_Bootstrap::onSicoCreditPlusFinancingStatusChanged()
	 * @param \Shopware\Models\Order\Order $oOrder
	 *
	 * @throws \Exception If a query builder function can not be found
	 */
	public function checkForUpdateMail($oOrder) {
		$oContractData = $this->getContractData($oOrder);
		/** @var \Shopware\Models\Order\Repository $oOrderRep */
		$oOrderRep = $this->getModelManager()->getRepository('Shopware\Models\Order\Order');
		if ( method_exists($oOrderRep, 'getBackendOrdersQueryBuilder') ) {
			$oQuery = $oOrderRep->getBackendOrdersQueryBuilder(array(array('property' => 'orders.id', 'value' => $oOrder->getId())))->getQuery();
		} elseif ( method_exists($oOrderRep, 'getOrdersQueryBuilder') ) {
			$oQuery = $oOrderRep->getOrdersQueryBuilder(array(array('property' => 'orders.id', 'value' => $oOrder->getId())))->getQuery();
		} else {
			throw new \Exception('No Order Query method available.');
		}
		$oQuery->setHydrationMode(\Doctrine\ORM\AbstractQuery::HYDRATE_ARRAY);
		$oPaginator = $this->getModelManager()->createPaginator($oQuery);
		$aOrders = $oPaginator->getIterator()->getArrayCopy();
		if ( $aOrders && isset($aOrders[0])) {
			$aOrder = $aOrders[0];
			$sNewState = $this->getCreditPlusPaymentState($oContractData->state, $oContractData->finallyApproved, $oContractData->deliveryDone, $oContractData->informationRequest);
			$oState = $this->getPaymentStateByName($sNewState);
			if ( $oState && ($aOrder['paymentStatus']['id'] != $oState->getId()) ) {
				if ( $sNewState != 'review_necessary' ) {
					// Update in the database
					// only if it's not "review_necessary"
					$oOrderHandler = Shopware()->Modules()->Order();
					$oOrderHandler->setPaymentStatus($aOrder['id'], $oState->getId());
					$oEventArgs = new \Enlight_Controller_EventArgs(array(
						'aOrder'  => $aOrder,
						'oOrder'  => $oOrder
					));
					Shopware()->Events()->notify(
						'SicoCreditPlus_Financing_Status_Changed',
						$oEventArgs
					);
				} else {
					/** @var Repository $oRQRep */
					$oRQRep = $this->getModelManager()->getRepository('SicoCreditPlus\Models\RequestQueue\RequestQueue');
					// Only add a new one if there are no open ones (like the currently requested)
					if ( !($aRQs = $oRQRep->findUnexecutedEntriesForOrder($oOrder)) || (count($aRQs) == 0) ) {
						$oRQ = new RequestQueue();
						// 40-80 minutes
						$iSeconds = rand(2400, 4800);
						try {
							$oDI = new \DateInterval( 'PT' . $iSeconds . 'S' );
							$oDate = new \DateTime();
							$oDate->add($oDI);
							$oRQ->setExecutionDate($oDate);
						} catch ( \Exception $oE ) {
							// The interval should always succeed
						}
						$oRQ->setOrder($oOrder)
							->setAlreadyExecuted(0)
							->setRetryRound(1);
						$oRQRep->append($oRQ);
					}
				}
			}
		}
	}

	/**
	 * @return \Shopware\Components\Model\ModelManager
	 */
	protected function getModelManager() {
		return Shopware()->Models();
	}
}
