<?php
/**
 * Created by PhpStorm.
 * @author mihovil.bubnjar
 * @date 11.12.2018
 * @time 11:31
 */

namespace SicoCreditPlus\Components;

use Shopware\Components\Logger;
use SicoCreditPlus\Lib\Controller\WSSoapClient;
use SicoCreditPlus\Lib\CreditPlusHelper\AbstractShopLogger;

class SicoCreditPlusLogger extends AbstractShopLogger {

	/**
	 * @var \Shopware\Components\DependencyInjection\Container|null
	 */
	private $oContainer = null;

	/**
	 * Write SoapClient Request-Info
	 * to Log-File
	 *
	 * @param string $sToday Timestamp for log file entry
	 * @param bool $bError Whether or not this file is prefixed with error
	 * @param string $sFunctionName Which function has been called
	 * @param string $sContent Content to log
	 */
	public function writeLastRequestInfo( $sToday = '', $bError = false, $sFunctionName = '', $sContent = '' ) {
		/** @var Logger $oLogger */
		$oLogger = $this->get('pluginlogger');
		$sContent = $this->cleanUpContent($sContent);
		if ( $bError ) {
			$oLogger->error($sContent, array('function' => $sFunctionName, 'date' => $sToday, 'plugin' => 'SicoCreditPlus'));
		} else {
			$oLogger->info($sContent, array('function' => $sFunctionName, 'date' => $sToday, 'plugin' => 'SicoCreditPlus'));
		}
	}

	/**
	 * Write SoapClient Request-Info
	 * to Log-File
	 *
	 * @param string $sToday Timestamp for log file entry
	 * @param bool $bError Whether or not to use error prefix in file name
	 * @param string $sFunctionName Which function was called
	 * @param string $sContent Content to log
	 */
	public function writeLastResponseInfo( $sToday = '', $bError = false, $sFunctionName = '', $sContent = '' ) {
		/** @var Logger $oLogger */
		$oLogger = $this->get('pluginlogger');
		$sContent = $this->cleanUpContent($sContent);
		if ( $bError ) {
			$oLogger->error($sContent, array('function' => $sFunctionName, 'date' => $sToday, 'plugin' => 'SicoCreditPlus'));
		} else {
			$oLogger->info($sContent, array('function' => $sFunctionName, 'date' => $sToday, 'plugin' => 'SicoCreditPlus'));
		}
	}

	/**
	 * Write Request-Information-Data
	 * to Log-File
	 *
	 * @param string $sToday Timestamp for log file entry
	 * @param string $sFunctionName The called API function name (e.g. getContracts)
	 * @param string[] $aArgumentData The arguments passed to $sFunctionName
	 * @param bool $bError true, if logged event is an error
	 */
	public function writeRequestInformation( $sToday = '', $bError = false, $sFunctionName = '', $aArgumentData = array() ) {
		/** @var Logger $oLogger */
		$oLogger = $this->get('pluginlogger');
		// Do a var_export, filter out new lines and replace them with exactly one space character
		if ( isset($aArgumentData) && isset($aArgumentData[1]) && ($aArgumentData[1] instanceof WSSoapClient) ) {
			//$aArgumentData[1] = $this->prepareVarExportSoapClient($aArgumentData[1]);
			$aArgumentData[1] = $this->dumpSoapClient($aArgumentData[1]);
		}
		$sArgumentData = var_export($aArgumentData, true);
		if( strlen($sArgumentData) > 5000000 ) { return; }
		$sArgumentData = $this->cleanUpContent($sArgumentData);
		$sContent = $sFunctionName.' '.$sArgumentData;
		if ( $bError ) {
			$oLogger->error($sContent, array('function' => $sFunctionName, 'date' => $sToday, 'plugin' => 'SicoCreditPlus'));
		} else {
			$oLogger->info($sContent, array('function' => $sFunctionName, 'date' => $sToday, 'plugin' => 'SicoCreditPlus'));
		}
	}

	/**
	 * Sets the Container.
	 *
	 * @param \Shopware\Components\DependencyInjection\Container|null $Container A ContainerInterface instance or null
	 */
	public function setContainer( $Container = null ) {
		$this->oContainer = $Container;
	}

	protected function get($sKey) {
		if ( $this->oContainer !== null ) {
			$oReturn = $this->oContainer->get($sKey);
			if ( substr($sKey,-6) == 'logger' ) {
				/** @var Logger $oReturn */
				if ( method_exists($oReturn,'getHandlers') ) {
					$oHandlers = $oReturn->getHandlers();
					foreach ( $oHandlers as $oHandler ) {
						if ( $oHandler instanceof \Monolog\Handler\RotatingFileHandler ) {
							$oHandler->setLevel(Logger::DEBUG);
						}
					}
				}
				// 4.3.x needs no correction as the default level is debug
			}
			return $oReturn;
		} else {
			if ( substr($sKey,-6) == 'logger' ) {
				$sLoggerType = substr($sKey,0, -6);
				// Document Root, kind of
				$sLogFile = dirname(dirname(dirname(dirname(dirname(dirname(dirname(__FILE__))))))).DIRECTORY_SEPARATOR;
				$sLogFile .= 'logs'.DIRECTORY_SEPARATOR.$sLoggerType.'_production.log';
				$oHandler = new \Monolog\Handler\RotatingFileHandler($sLogFile, 14);
				$oHandler->pushProcessor(new \Monolog\Processor\UidProcessor());
				$oLogger = new Logger($sLoggerType);
				$oLogger->pushHandler($oHandler);
				return $oLogger;
			}
			return null;
		}
	}

	/**
	 * Removes __soap_fault from WSSoapClient to lessen the burden on var_export
	 *
	 * @param \SicoCreditPlus\Lib\Controller\WSSoapClient $oSoapClient
	 *
	 * @return \SicoCreditPlus\Lib\Controller\WSSoapClient
	 */
	private function prepareVarExportSoapClient( $oSoapClient ) {
		if ( isset($oSoapClient->__soap_fault) && isset($oSoapClient->__soap_fault->trace) ) {
			$oSoapClient->__soap_fault = null;
		}
		return $oSoapClient;
	}

	private function dumpSoapClient(WSSoapClient $oSoapClient){
		$ret = [];
		//$ret['Cookies'] = $oSoapClient->__getCookies();
		$ret['lastRequest'] = $oSoapClient->__getLastRequest();
		$ret['lastRequestHeaders'] = $oSoapClient->__getLastRequestHeaders();
		$ret['lastResponse'] = $oSoapClient->__getLastResponse();
		$ret['lastResponseHeaders'] = $oSoapClient->__getLastResponseHeaders();
		return $ret;
	}
}
