/**
 * Shopware 5
 * Copyright (c) shopware AG
 *
 * According to our dual licensing model, this program can be used either
 * under the terms of the GNU Affero General Public License, version 3,
 * or under a proprietary license.
 *
 * The texts of the GNU Affero General Public License with an additional
 * permission and of our proprietary license can be found at and
 * in the LICENSE file you have received along with this program.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * "Shopware" is a registered trademark of shopware AG.
 * The licensing of the program under the AGPLv3 does not imply a
 * trademark license. Therefore any rights, title and interest in
 * our trademarks remain entirely with us.
 *
 * @category   Shopware
 * @package    SicoCreditPlusProductgroups
 * @subpackage Settings
 * @version    $Id$
 * @author SinkaCom AG
 */

/* {namespace name="frontend/plugins/payment/creditplus/backend"} */

/**
 * Shopware UI - Productgroup Article Mapping
 *
 * Shows the drag and drop selector to map articles to a productgroup
 */
//{block name="backend/sico_credit_plus_productgroups/view/detail/article_assignment"}
Ext.define('Shopware.apps.SicoCreditPlusProductgroups.view.detail.ArticleAssignment', {

   /**
    * Parent Element Ext.container.Container
    * @string
    */
    extend:'Ext.form.Panel',

    /**
     * Register the alias for this class.
     * @string
     */
    alias:'widget.sicocreditplus-productgroups-tabs-article_assignment',

    /**
     * Base class of the component
     * @string
     */
    cls: 'shopware-form',

    /**
     * Specifies the border for this component. The border can be a single numeric
     * value to apply to all sides or it can be a CSS style specification for each
     * style, for example: '10 5 3 10'.
     *
     * Default: 0
     * @integer
     */
    border: 0,

    /**
     * Display the the contents of this tab immediately
     * @boolean
     */
    autoShow: true,

    /**
     * Layout configuration
     * @object
     */
    layout: {
        type: 'hbox',
        pack: 'start',
        align: 'stretch'
    },

    /**
     * Body padding
     * @integer
     */
    bodyPadding: 10,

    /**
     * Available action buttons
     * @array
     */
    defaultButtons: [ 'add', 'remove' ],

    /**
     * Default text which are used for the tooltip on the button.
     * @object
     */
    buttonsText: {
        add: '{s name="Add"}Hinzufügen{/s}',
        remove: '{s name="Remove"}Entfernen{/s}'
    },


    /**
     * Initialize the Shopware.apps.SicoCreditPlusProductgroups.view.detail.ArticleAssignment and defines the necessary
     * default configuration
     *
     * @return { void }
     */
    initComponent: function () {
        var me = this;
        me.fromGrid = me.createFromGrid();
        me.buttonContainer = me.createActionButtons();
        me.toGrid = me.createToGrid();

        me.items = [ me.fromGrid, me.buttonContainer, me.toGrid ];
        me.addEvents('storeschanged', 'add', 'remove');
        me.on('storeschanged', me.onStoresChanged, me);

        me.callParent(arguments);
    },

    /**
     * Creates the `from` grid
     * @return { Ext.grid.Panel }
     */
    createFromGrid: function() {
		var me = this;
		return me.createStandardGrid(me.availableProductsStore, '{s name="AvailableArticles"}Verfügbare Artikel{/s}', 'from');
	},

    /**
     * Creates the `to` grid
     * @return { Ext.grid.Panel }
     */
    createToGrid: function() {
		var me = this;
		return me.createStandardGrid(me.assignedProductsStore, '{s name="MappedArticles"}Zugewiesene Artikel{/s}', 'to');
	},

	/**
	 * Creates any grid with three parameters
	 * @param { Ext.data.Store } store
	 * @param { String } title
	 * @param { String } internalTitle
	 * @return { Ext.grid.Panel }
	 */
	createStandardGrid: function(store, title, internalTitle) {
		var me = this, grid, toolbar;

		grid = Ext.create('Ext.grid.Panel', {
			internalTitle: internalTitle,
			title: title,
			flex: 1,
			store: store.load(),
			selModel: me.createSelectionModel(),
			viewConfig: {
				loadMask: false,
				plugins: me.createGridDragAndDrop(),
				listeners: {
					drop: me.onDragAndDropAssignment
				}
			},
			sicoView: me,
			bbar: me.createPagingToolbar(store),
			columns: me.getColumns()
		});

		toolbar = me.createSearchToolbar(grid);
		grid.addDocked(toolbar);

		return grid;
	},

    /**
     * Creates the action buttons which are located between the `fromGrid` (on the left side)
     * and the `toGrid` (on the right side).
     *
     * The buttons are placed in an `Ext.container.Container` to apply the necessary layout
     * on it.
     *
     * @return { Ext.container.Container }
     */
    createActionButtons: function() {
        var me = this;

        me.actionButtons = [];
        Ext.Array.forEach(me.defaultButtons, function(name) {

            var button = Ext.create('Ext.Button', {
                tooltip: me.buttonsText[name],
                cls: Ext.baseCSSPrefix + 'form-itemselector-btn',
                iconCls: Ext.baseCSSPrefix + 'form-itemselector-' + name,
                action: name,
                disabled: true,
                navBtn: true,
                margin: '4 0 0 0',
                listeners: {
                    scope: me,
                    click: function() {
                        me.fireEvent(name, me);
						if ( name == 'add' ) {
							me.onAddProducts(me);
						} else if ( name == 'remove' ) {
							me.onRemoveProducts(me);
						}
                    }
                }
            });
            me.actionButtons.push(button);
        });


        return Ext.create('Ext.container.Container', {
            margins: '0 4',
            items:  me.actionButtons,
            width: 22,
            layout: {
                type: 'vbox',
                pack: 'center'
            }
        });
    },

    /**
     * Creates a paging toolbar based of the incoming store
     *
     * @param { Ext.data.Store } store
     * @return { Ext.toolbar.Paging }
     */
    createPagingToolbar: function(store) {

        return Ext.create('Ext.toolbar.Paging', {
            store: store,
            displayInfo: true
        });
    },

    /**
     * Creates a toolbar which could be docked to the top of
     * a grid panel and contains a searchfield to filter
     * the associated grid panel.
     *
     * @return { Ext.toolbar.Toolbar }
     */
    createSearchToolbar: function(cmp) {
        var me = this, searchField, toolbarItems = [], iSearchFieldWidth = 270;

        
		
		if ( cmp.internalTitle == 'from' ) {
			var assignAllButton = Ext.create('Ext.button.Button', {
				name: 'assignall',
				dock: 'top',
				width: 120,
				text: '{s name="ButtonAssignAll"}Alle zuweisen{/s}',
				handler: function() {
					var url = '{url controller="SicoCreditPlusProductgroups" action="assignAllArticles"}';
					var message = '{s name="AddAllSuccess"}All articles assigned{/s}';
					var failure = '{s name="AddAllError"}An error occured assigning all articles{/s}';
					me.setLoading(true);

					Ext.Ajax.request({
						url: url,
						params: { id: parseInt(me.record.get('id')) },
						success: function(response) {

							var result = Ext.decode(response.responseText);
							message = Ext.String.format(message, result.counter);
							Shopware.Notification.createGrowlMessage('',message);

							//reload the stores for the paging bar information
							me.toGrid.getStore().load({
								callback:function () {
									me.setLoading(false);
								}
							});
							me.fromGrid.getStore().load();
							me.record.getArticlesStore = me.toGrid.getStore();

						},
						failure: function(response) {
							me.setLoading(false);

							var result = Ext.decode(response.responseText);
							failure = failure + '<br>' + result.error;
							Shopware.Notification.createGrowlMessage('',message);
						}
					});
				}
			});
			toolbarItems.push(assignAllButton);
			iSearchFieldWidth = 220;
		}
		toolbarItems.push('->');

		searchField = Ext.create('Ext.form.field.Text', {
			name: 'searchfield',
			dock: 'top',
			cls: 'searchfield',
			width: iSearchFieldWidth,
			emptyText: 'Search...',
			enableKeyEvents: true,
			checkChangeBuffer: 500,
			listeners: {
				change: function(field, value) {
					me.fireEvent('search', value, cmp);
					me.onSearch(value, cmp);
				}
			}
		});

		toolbarItems.push(searchField);
		toolbarItems.push(' ');
        return Ext.create('Ext.toolbar.Toolbar', {
            ui: 'shopware-ui',
            padding: '2 0',
            items: toolbarItems
        });
    },

    /**
     * Creates the selection model which is used by both grids.
     *
     * @returns { Ext.selection.CheckboxModel }
     */
    createSelectionModel: function() {
		var me = this;
        return Ext.create('Ext.selection.CheckboxModel',{
			mode: 'SIMPLE',
			listeners: {
				selectionchange: me.onSelectionChange
			},
			sicoView: me
		});
    },

    createGridDragAndDrop: function() {
        return Ext.create('Ext.grid.plugin.DragDrop', {
            ddGroup: 'sicocreditplus-productgroups-article-assignment-grid-dd'
        });
    },

    /**
     * Creates the necessary columns for both grids. Please
     * note that the `name` column has a specific renderer.
     *
     * @return { Array }
     */
    getColumns: function() {
        var me = this;

        return [{
            header: '{s name="ArticleNumber"}Article Number{/s}',
            flex: 1,
            dataIndex: 'number'
        }, {
            header: '{s name="ArticleName"}Article Name{/s}',
            flex: 2,
            dataIndex: 'name',
            renderer: me.nameColumnRenderer
        }, {
            header: '{s name="SupplierName"}Supplier Name{/s}',
            flex: 1,
            dataIndex: 'supplierName'
        }];
    },

    /**
     * Renders the incoming column value into `strong` tags.
     *
     * @param { String } value
     * @return { String } formatted string
     */
    nameColumnRenderer: function(value) {
        return Ext.String.format('<strong>[0]</strong>', value);
    },

	/**
	 * Triggers a search by using an `extraParam` on the
	 * associated store of the active grid.
	 *
	 * @param { String } value
	 * @param { Ext.grid.Panel } activeGrid
	 */
	onSearch: function(value, activeGrid) {
		var store = activeGrid.getStore();

		value = Ext.String.trim(value);
		store.currentPage = 1;

		store.getProxy().extraParams.search = (!value.length) ? '' : value;
		store.load();
	},

    /**
     * Event listener which will be fired when the user selects
     * an another productgroup in the list.
     *
     * The method reconfigures the stores and reloads them.
     *
     * @return { void }
     */
    onStoresChanged: function() {
        var me = this,
            fromStore = me.availableProductsStore,
            toStore = me.assignedProductsStore;

        // Set the new stores
        me.fromGrid.reconfigure(fromStore);
        me.toGrid.reconfigure(toStore);

        // Reload the stores
        me.fromGrid.getStore().load();
        me.toGrid.getStore().load();
    },

	/**
	 * Event listener method which will be fired when the user presses
	 * the upper button in the middle of the component, if the user
	 * has a selection.
	 *
	 * The method collects the id's of the selected records and adds
	 * them to the correct grid.
	 *
	 * @param { Shopware.apps.SicoCreditPlusProductgroups.view.detail.ArticleAssignment } scope
	 * @returns { Boolean }
	 */
	onAddProducts: function(scope) {
		var activeGrid = scope.fromGrid,
			inactiveGrid = scope.toGrid,
			store = activeGrid.getStore(),
			inactiveStore = inactiveGrid.getStore(),
			selection = activeGrid.getSelectionModel().getSelection(),
			ids = [], productgroupId;

		if(!selection.length) {
			return false;
		}

		Ext.each(selection, function(sel) {
			ids.push(sel.data.articleId);
		});

		store.remove(selection);
		inactiveStore.add(selection);

		productgroupId = store.getProxy().extraParams.id;

		scope._sendRequest('add', ids, productgroupId);
		return true;
	},

	/**
	 * Event listener method which will be fired when the user presses
	 * the lower button in the middle of the component, if the user
	 * has a selection.
	 *
	 * The method collects the id's of the selected records and adds
	 * them to the correct grid.
	 *
	 * @param { Shopware.apps.SicoCreditPlusProductgroups.view.detail.ArticleAssignment } scope
	 * @returns { Boolean }
	 */
	onRemoveProducts: function(scope) {
		var activeGrid = scope.toGrid,
			inactiveGrid = scope.fromGrid,
			store = activeGrid.getStore(),
			inactiveStore = inactiveGrid.getStore(),
			selection = activeGrid.getSelectionModel().getSelection(),
			ids = [], productgroupId;

		if(!selection.length) {
			return false;
		}

		Ext.each(selection, function(sel) {
			ids.push(sel.data.articleId);
		});

		store.remove(selection);
		inactiveStore.add(selection);

		productgroupId = store.getProxy().extraParams.id;

		scope._sendRequest('remove', ids, productgroupId);
		return true;
	},

	/**
	 * Event listener method which will be fired when the user drags
	 * records from one grid to the other one.
	 *
	 * The method collects the id's of the dropped records.
	 *
	 * @param { HTMLElement } node
	 * @param { Object } data
	 * @returns { void }
	 */
	onDragAndDropAssignment: function(node, data) {
		var activeView = data.view,
			activeGrid = activeView.panel,
			me = activeGrid.sicoView,
			store = activeGrid.getStore(),
			records = data.records, action, productgroupId, ids = [];

		action = (activeGrid.internalTitle === 'from') ? 'add' : 'remove';

		Ext.each(records, function(record) {
			ids.push(record.data.articleId);
		});

		productgroupId = store.getProxy().extraParams.id;

		me._sendRequest(action, ids, productgroupId);
	},

	/**
	 * Helper method which sents the AJAX request to add / remove
	 * the records, which are associated with the incoming id's.
	 *
	 * @param { String } action - Action which will be used for the request: add (default), remove
	 * @param { Array } ids - Array of record id's
	 * @param { integer } productgroupId - Id of the selected Productgroup
	 * @private
	 */
	_sendRequest: function(action, ids, productgroupId) {
		var mapping = this;
		var url = '{url controller="SicoCreditPlusProductgroups" action="addAssignedArticles"}';
		var message = '{s name="AddSuccess"}[0]x articles assigned{/s}';
		var failure = '{s name="AddError"}The following error occurred while adding the articles:{/s}';

		if(action === 'remove') {
			url = '{url controller="SicoCreditPlusProductgroups" action="removeAssignedArticles"}';
			message = '{s name="RemoveSuccess"}[0]x articles assignments removed{/s}';
			failure = '{s name="RemoveError"}The following error occurred while removing the articles:{/s}';
		}
		mapping.setLoading(true);

		Ext.Ajax.request({
			url: url,
			params: { ids: Ext.JSON.encode(ids), id: parseInt(productgroupId) },
			success: function(response) {

				var result = Ext.decode(response.responseText);
				message = Ext.String.format(message, result.counter);
				Shopware.Notification.createGrowlMessage('',message);

				//reload the stores for the paging bar information
				mapping.toGrid.getStore().load({
					callback:function () {
						mapping.setLoading(false);
					}
				});
				mapping.fromGrid.getStore().load();
				mapping.record.getArticlesStore = mapping.toGrid.getStore();

			},
			failure: function(response) {
				mapping.setLoading(false);

				var result = Ext.decode(response.responseText);
				failure = failure + '<br>' + result.error;
				Shopware.Notification.createGrowlMessage('',message);
			}
		});
	},

	/**
	 * Event listener method which will be fired when the user selects an entry
	 * in either the `toGrid` or the `fromGrid`.
	 *
	 * The method handles the enables / disables the buttons in the middle column
	 * of the view and deselects all entries in the inactive grid.
	 *
	 * @param { Ext.selection.CheckboxModel } selModel
	 * @param { Array } selection
	 * @returns { boolean }
	 */
	onSelectionChange: function(selModel, selection) {
		var me = this.sicoView,
			activeGrid = selModel.view.panel, inactiveGrid,
			activeBtn;

		// Prevent the grid to get a little sluggish
		if(!selection.length) {
			return false;
		}

		inactiveGrid = (activeGrid.internalTitle === 'from') ? me.toGrid : me.fromGrid;
		if ( activeGrid.internalTitle === 'from' ) {
			activeBtn = 'add';
		} else {
			activeBtn = 'remove';
		}
		Ext.each(me.actionButtons, function (oButton) {
			if ( oButton.action == activeBtn ) {
				oButton.setDisabled(false);
			} else {
				oButton.setDisabled(true);
			}
		});

		inactiveGrid.getSelectionModel().deselectAll(true);

		return true;
	}
});
//{/block}
